from autobahn.websocket.protocol import WebSocketClientProtocol as AbstractProtocol
from autobahn.twisted import WebSocketClientProtocol, WebSocketClientFactory

from .tcp import makeTCPClient


def makeWebSocketClient(context, server, protocol=None,
                        factory=None, factoryClass=None, endpoint='', close='sendClose'):
    if factory is None:
        url = "ws://{}:{}{}".format(server.targetHost, server.targetPort, endpoint)
        if factoryClass is None:
            factoryClass = WebSocketClientFactory
        factory = factoryClass(url)
    if protocol is None:
        protocol = factory.protocol
        if protocol is AbstractProtocol:
            # bug in autobahn?
            protocol = WebSocketClientProtocol
    return makeTCPClient(context, protocol, server, factory, when='onOpen', close=close)



from __future__ import print_function

from os import environ

from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks

DEFAULT_TIMEOUT = environ.get('CARLY_DEFAULT_TIMEOUT', 0.2)


def withTimeout(deferred, timeout=None):
    minimum = environ.get('CARLY_MINIMUM_TIMEOUT')
    minimum = float(minimum) if minimum else None
    timeout = timeout or minimum or DEFAULT_TIMEOUT
    if minimum is not None:
        timeout = max(timeout, minimum)
    return deferred.addTimeout(timeout, reactor)


def cancelDelayedCalls(expected=2):
    """
    :param expected:
      The number of calls to cancel. If the number found does not match this,
      none of them will be cancelled so that trial's cleanup can tell you
      more about them.

      Why the default of 2? Hopefully you're only testing one delayed calll
      generator at a time, and there's one for trial's 2 minute timeout.
    """
    calls = reactor.getDelayedCalls()
    strings = []
    for call in calls:
        strings.append(str(call))
        call.cancel()
    if len(calls) != expected:
        raise AssertionError(
            '\n\nExpected {} delayed calls, found {}: {}'.format(
                expected, len(calls), strings
        ))


def _pump():
    d = Deferred()
    reactor.callLater(0, lambda: d.callback(None))
    return d


@inlineCallbacks
def advanceTime(seconds):
    """
    Advance the reactor time by the number of seconds or partial seconds
    specified.
    """
    yield _pump()
    now = reactor.seconds()
    for call in reactor.getDelayedCalls():
        currentSecondsFromNow = call.getTime() - now
        newSecondsFromNow = max(0, currentSecondsFromNow - seconds)
        call.reset(newSecondsFromNow)
    # give the reactor a chance to run calls we're brought forward:
    yield _pump()



from attr import make_class
from twisted.internet.defer import (
    inlineCallbacks, gatherResults, returnValue
)
from twisted.internet.protocol import Factory, ClientFactory

from .hook import hook

TCPClient = make_class('TCPClient', ['protocolClass', 'clientProtocol', 'serverProtocol'])


class TCPServer(object):

    def __init__(self, protocolClass, port):
        self.protocolClass = protocolClass
        self.port = port
        host = self.port.getHost()
        self.targetHost = host.host
        self.targetPort = host.port


def makeTCPServer(context, protocol, factory=None, interface='127.0.0.1',
                  installProtocol=True):

    from twisted.internet import reactor

    hook(protocol, 'connectionMade')
    if factory is None:
        factory = Factory()
    if installProtocol:
        factory.protocol = protocol
    port = reactor.listenTCP(0, factory, interface=interface)
    server = TCPServer(protocol, port)
    hook(server.protocolClass, 'connectionLost', once=True)
    context.cleanupServers(server.port)
    return server


def disconnect(client):
    client.clientProtocol.transport.loseConnection()


def makeTCPClient(context, protocol, server, factory=None,
                  when='connectionMade', close=disconnect):

    from twisted.internet import reactor

    hook(protocol, when)
    if factory is None:
        factory = ClientFactory()
    factory.protocol = protocol
    host = server.port.getHost()
    reactor.connectTCP(host.host, host.port, factory)
    return waitForClient(
        context, getattr(protocol, when), server.protocolClass.connectionMade, close
    )


@inlineCallbacks
def waitForClient(context, clientConnected, serverConnected, close=disconnect):
    clientProtocol, serverProtocol = yield gatherResults([
        clientConnected.protocol(), serverConnected.protocol(),
    ])
    client = TCPClient(clientProtocol.__class__, clientProtocol, serverProtocol)
    hook(client.protocolClass, 'connectionLost', once=True)
    context.cleanupClient(client, close)
    returnValue(client)



from time import sleep

from twisted.internet import reactor
from twisted.internet.threads import deferToThread

from .clock import withTimeout


def pendingIsEmpty():
    while True:
        stats = reactor.threadpool._team.statistics()
        if not (stats.backloggedWorkCount or stats.busyWorkerCount > 1):
            break
        sleep(0.001)


def waitForThreads(timeout=None):
    return withTimeout(deferToThread(pendingIsEmpty), timeout)



from functools import partial

from twisted.internet.defer import (
    inlineCallbacks, gatherResults, maybeDeferred
)

from .clock import withTimeout, cancelDelayedCalls
from .hook import cleanup
from .threads import waitForThreads


class Context(object):

    def __init__(self):
        self.cleanups = {
            'connections': [],
            'listens': [],
        }

    def _cleanup(self, cleanups, timeout):
        deferreds = []
        for p in cleanups:
            d = p()
            deferreds.append(d)
            withTimeout(d, timeout)
        return gatherResults(deferreds)

    @inlineCallbacks
    def cleanup(self, timeout=None, threads=False, delayedCalls=None):
        yield self._cleanup(self.cleanups['connections'], timeout)
        yield self._cleanup(self.cleanups['listens'], timeout)
        cleanup()
        if threads:
            yield waitForThreads()
        if delayedCalls:
            cancelDelayedCalls(delayedCalls)

    def cleanupServers(self, *ports):
        self.cleanups['listens'].extend(
            partial(maybeDeferred, port.stopListening) for port in ports
        )

    def cleanupClient(self, client, close, timeout=None):
        if isinstance(close, str):
            name = close
            close = lambda client: getattr(client.clientProtocol, name)()
        self.cleanups['connections'].extend((
            partial(maybeDeferred, close, client),
            partial(client.clientProtocol.connectionLost.called, timeout=timeout),
            partial(client.serverProtocol.connectionLost.called, timeout=timeout),
        ))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['carly']

package_data = \
{'': ['*']}

install_requires = \
['Twisted>=18', 'attrs']

setup_kwargs = {
    'name': 'carly',
    'version': '0.11.0',
    'description': 'A tool for putting messages into and collecting responses from Twisted servers using real networking',
    'long_description': 'carly\n=====\n\n|CircleCI|_\n\n.. |CircleCI| image:: https://circleci.com/gh/cjw296/carly/tree/master.svg?style=shield\n.. _CircleCI: https://circleci.com/gh/cjw296/carly/tree/master\n\n\nA tool for putting messages into and collecting responses from Twisted servers and clients using real networking.\n\n"call me maybe"\n\nWhy carly? \'cos someone already took `Jepsen!`__\n\n__ https://jepsen.io/\n',
    'author': 'Chris Withers',
    'author_email': 'chris@withers.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cjw296/carly',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*',
}


setup(**setup_kwargs)



from twisted.internet.protocol import DatagramProtocol

from .hook import hook


class UDP(DatagramProtocol):

    def __init__(self, port, protocol):
        self.port = port
        self.protocol = protocol
        host = self.port.getHost()
        self.targetHost = host.host
        self.targetPort = host.port

    def startProtocol(self):
        self.transport.connect(self.targetHost, self.targetPort)

    def send(self, datagram):
        self.transport.write(datagram)


def makeUDP(context, protocol, interface='127.0.0.1'):

    from twisted.internet import reactor

    hook(protocol.__class__, 'datagramReceived')
    port = reactor.listenUDP(0, protocol, interface)
    udp = UDP(port, protocol)
    sendPort = reactor.listenUDP(0, udp, interface)
    context.cleanupServers(port, sendPort)
    return udp




from __future__ import print_function

from .clock import cancelDelayedCalls, advanceTime
from .context import Context
from .hook import hook, cleanup, decoder, register
from .threads import waitForThreads
